% buildSymmetricalTemplate.m
% Developed By: Kenneth I. Vaden, Jr. and Mark A. Eckert
% Edited: Mark A. Eckert, 2/25/2019

% SUMMARY: This MATLAB script creates a study-specific template % using parallelized code. 
% Thus, you can start multiple MATLAB sessions and run this script.

% Your image directory(ies) should include the left-right flipped version of each image 
% to create a symmetrical template, as in PMCID ...
% The SPM Re-orient Utility can be used to create left-right flipped versions of each image. 

% PARAMETERS, 10 successive average images templates are % created, plus the final template, 
% after each normlization step % to refine the spatial resolution of the template or anatomical 
% target space for normalization.

% Please note that we have indicated below where you should provide your 
% project-specific information with % EDIT HERE.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

maxstep = 10; % Final-warping step (+1 for final average template)
wmins = 5;   % How many minutes to wait before checking for all finished files

% subject list
sublist = '/yourPath/IDs.txt';                                   % EDIT HERE.
Isbj = textread(sublist,'%s','delimiter','\n');
Inum = length(Isbj);

% Input images
Idir = '/yourPath/images/';                                      % EDIT HERE. input dir.
Iprf = '';                                                       % EDIT HERE. add prefix, if necessary (e.g., 'r' for rigidly aligned images) 
Iext = '.nii';                                                   % EDIT HERE. extension after subject ID  

% Input extension
Iname = fullfile(Idir,sprintf('%s%%s%s',Iprf,Iext));  % input name

% Check that the images exist
if (Inum==0); error('subject list is empty or does not exist!'); end;
sl=length(Isbj);
Fnum = 0; for (snum = 1:sl); if(exist(sprintf(Iname,Isbj{snum}))); Fnum = Fnum + 1; end; end;
if (Fnum ~= Inum); error(sprintf('number of images (%d) does not match list (%d)!',Fnum,Inum)); else fprintf('Found all %d images\n',Fnum); end;

% FOR EACH STEP
for cstep = 1:maxstep+1; % current step
    
    % TARGET FILENAME
    if (cstep <= maxstep);   tgt = fullfile(Idir,sprintf('symAverage_Step%02d.nii',cstep)); end;
    if (cstep == maxstep+1); tgt = fullfile(Idir,sprintf('symAverage_Final_Step%02d.nii',cstep)); end;
    
    % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
    % FIRST -- Create a target for normaliation by averaging (and deleting previous out-files from the previous step) if the current target does not yet exist
    if (~exist(tgt,'file'));
        % Create average target
        if (cstep == 1);
            inpt = fullfile(Idir,sprintf('%s*.nii',Iprf));                          % EDIT HERE.
            cmmd = sprintf('AverageImages 3 %s 0 %s',tgt,inpt);
            fprintf('%s\n',cmmd); unix(cmmd); disp(' ');

            % Create average target & erase prev step output
        elseif ((cstep > 1) && (cstep <= maxstep));
            inpt = fullfile(Idir,sprintf('OUTT%d_%s*.nii',cstep-1,Iprf));
            cmmd = sprintf('AverageImages 3 %s 0 %s',tgt,inpt);
            fprintf('%s\n',cmmd); unix(cmmd); disp(' ');

            % Create final average template & erase out-files from steps 1:(max-1)
        elseif (cstep == maxstep+1);
            tgt = fullfile(Idir,sprintf('symAverage_Final_Step%02d.nii',cstep));    
            inpt = fullfile(Idir,sprintf('OUTT%d_%s*.nii',cstep-1,Iprf));           
            cmmd = sprintf('AverageImages 3 %s 0 %s',tgt,inpt);
            fprintf('%s\n',cmmd); unix(cmmd); disp(' ');

            % Erase all but the final OUT* images
            fprintf('\nErasing all files, except for final step (%d)\n',maxstep-1);
            for ci = 1:(maxstep-1);
                cmmd = sprintf('rm %s', fullfile(Idir,sprintf('OUTT%d_*.*',ci)));   
                fprintf('%s\n',cmmd); unix(cmmd); disp(' ');
                cmmd = sprintf('rm %s', fullfile(Idir,sprintf('OUT%d_*.*',ci)));    
                fprintf('%s\n',cmmd); unix(cmmd); disp(' ');
            end;
        end;
    else
        fprintf('file already exists: %s, average target not (re-)generated.',tgt);
    end
    
    % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
    % NEXT, normalize the next image that has not yet been normalized in the current step 
    %   Filname for next-step average - check to make sure this step needs to be
    %   performed by the current cpu/matlab session
    if (cstep+1 <= maxstep);   Ntgt = fullfile(Idir,sprintf('symAverage_Step%02d.nii',cstep+1)); end;
    if (cstep+1 == maxstep+1); Ntgt = fullfile(Idir,sprintf('symAverage_Final_Step%02d.nii',cstep+1)); end;
    if ((cstep <= maxstep)&&(~exist(Ntgt,'file')));
        for s = 1:Inum; %for (snum = Isbj')

            % Affine/placeholder name (OUT_dr1241Affine.txt)
            Oname = fullfile(Idir,sprintf('OUT%d_%s%sAffine.txt',cstep,Iprf,Isbj{s}));

            % If the file does not already exist ...
            if (~exist(Oname,'file'));
                % write a textfile-placeholder
                dlmwrite(Oname,0,'delimiter','\t');
                fprintf('\n\n********************************************************\n');
                fprintf('  >> Process %d - placeholder created: %s\n',Isbj{s},Oname);
                fprintf('********************************************************\n\n');
                
                % Estimate warps to transform current individual native/warped brain 
                % to more closely approximate the group average
                inpf = fullfile(Idir,sprintf('%s%s.nii',Iprf,Isbj{s}));                    % EDIT HERE.
                MM = sprintf('-m PR[%s,%s,1,4]',tgt,inpf);
                OO = sprintf('-o %s/OUT%d_%s%s.nii',Idir, cstep,Iprf,Isbj{s});
                RR = '-r Gauss[4,1]'; TT = '-t SyN[2,1,1]'; GG = '--geodesic 1'; 
                
		% Changing warping parameters 
                if (cstep <  maxstep); II = '-i 30x90x20x12'; end;
                if (cstep == maxstep); II = '-i 30x90x30x20'; end;

                % Use sprintf to create a command that is passed to unix 
                clear cmmd; cmmd = sprintf('time ANTS 3 %s %s %s %s %s %s',MM,OO,II,RR,TT,GG);
                fprintf('%s\n',cmmd); unix(cmmd); disp(' ');
                
                % Apply the warping or normalization parameters
                AF = fullfile(Idir,sprintf('OUT%d_%s%sAffine.txt',cstep,Iprf,Isbj{s}));
                DF = fullfile(Idir,sprintf('OUT%d_%s%sWarp.nii',cstep,Iprf,Isbj{s}));
                XFORMS=sprintf('%s %s',DF,AF);

                % Output name has an extra T (OUTT) to avoid averaging warps into template
                OUTN = fullfile(Idir,sprintf('OUTT%d_%s%s.nii',cstep,Iprf,Isbj{s}));

                % 3D implement warps -> create individual normalized anat
                clear cmmd; cmmd = sprintf('WarpImageMultiTransform 3 %s %s -R %s %s',inpf,OUTN,tgt,XFORMS);
                fprintf('%s\n',cmmd); unix(cmmd); disp(' ');
                
            else
                fprintf('  >> Skip %s - file exists: %s\n',Isbj{s},Oname)
            end
        end
        
        % Before advancing to the next step, check finished files and next-step
        % average/target image. If not found, wait a couple minutes and then re-test
        Fnum = 0;
        while (Fnum < Inum);
            Fnum = 0;
            for s = 1:Inum; % Check whether each output file exists
                if exist(fullfile(Idir,sprintf('OUTT%d_%s%s.nii',cstep,Iprf,Isbj{s})),'file'); Fnum = Fnum + 1; end;
            end

            % Check to see whether next-step average file exists
            if (cstep+1 <= maxstep);   tgt = fullfile(Idir,sprintf('symAverage_Step%02d.nii',cstep+1)); end;
            if (cstep+1 == maxstep+1); tgt = fullfile(Idir,sprintf('symAverage_Final_Step%02d.nii',cstep+1)); end;
            if exist(tgt,'file'); Fnum = Inum; end;

            % If all finished files OR next-average was not found - wait
            if (Fnum < Inum);
                fprintf('\n %d / % Finished files. Waiting %d minutes to check finished files again ... \n',Fnum,Inum,wmins);
                pause(wmins*60); % Pause for 5 minutes before checking again
            end
        end
    end
    
    % Announce the conclusion of each step
    fprintf('\n\n************************************************************************');
    if (cstep <  maxstep);   fprintf('\n STEP %d FINISHED ... \n',cstep); end;
    if (cstep == maxstep);   fprintf('\n FINAL WARPING (STEP %d) FINISHED.\n',cstep); end;
    if (cstep == maxstep+1); fprintf('\n FINAL TEMPLATE GENERATED. \n'); end;
    fprintf('************************************************************************\n\n');
 
    
end
