# spatialResolution_FFT.R
# Developed By: Kenneth I. Vaden, Jr. and Mark A. Eckert
# Edited: Mark A. Eckert, 2/25/2019

# Evaluate the spatial resolution of 3D images.  
# This code was used to evaluate the spatial 
# resolution of Jacobian determinant images from 
# each warping step step in PMCID ...

# Global, local, and fine terms below refer to Step 1, Step 2, and Step 3 
# normalization iterations with increasingly finer resolution, respectively.

# Please note that the functions call library("AnalyzeFMRI")
# Please note that we have indicated below where you should provide your 
# project-specific information with # EDIT HERE.

MAINdir <- '/yourPath';                                                  # EDIT HERE
DSTdir <- sprintf('%s/spatialResol',MAINdir); # Script and results dir   # EDIT HERE
#  setwd(DSTdir); source('spatialResolution_FFT.R');

# OPERATING PARAMETERS: Calculate or Export to NIFTI
debugging <- 0; # Only import one image/sheet/etc for coding             # EDIT HERE to test first
masking   <- T; # Use anatomical image to mask jacobians
calc_stats   <- 1; # Step 1. Calculate stats for each slice, save freq info to txt
expt_results <- 0; # Step 2. Combine results across subjects

#################################################################
# Load R-packages
library("AnalyzeFMRI"); # nifti operations

# Import subject list
GList <- as.character(read.table(file = sprintf('%s/globalJacobianPathImageFilenames.txt',DSTdir),header = F,sep = '\t')$V1); #        "             "         "       "      (Step 1 files) # EDIT HERE
LList <- as.character(read.table(file = sprintf('%s/localJacobianPathImagesFilenames.txt',DSTdir),header = F,sep = '\t')$V1); #        "             "         "       "      (Step 2 files) # EDIT HERE
FList <- as.character(read.table(file = sprintf('%s/fineJacobianPathImagesFilenames.txt',DSTdir),header = F,sep = '\t')$V1);  # text file with paths and names to image files (Step 3 files) # EDIT HERE

# Testing on the local level only (LN=2)
LNS <- 1:3; if (debugging==1) { LNS <- 2; FList <- FList[1]; }
for (LN in LNS) {
  
  # Create directory for fine, local, global results:
  flgstr <- c('SpFine','SpLocal','SpGlobal')[LN]
  outpdir <- sprintf('%s/%s', DSTdir,flgstr);
  if (file.exists(outpdir)==0) {
    dir.create(outpdir, showWarnings=T, recursive=F, mode="0777");
    setwd(outpdir); writeLines(sprintf('%s  Created directory for results:\n   %s',Sys.time(),outpdir));
  }
  # Create directory for placeholder & intermed files
  tmpdir <- sprintf('%s/TMP', outpdir);
  if (file.exists(tmpdir)==0) {
    dir.create(tmpdir, showWarnings=T, recursive=F, mode="0777"); 
    setwd(tmpdir); writeLines(sprintf('%s  Created directory for results:\n   %s',Sys.time(),outpdir));
  }
  
  # File list & image directory
  if (LN == 1) { FL <- FList; }
  if (LN == 2) { FL <- LList; }
  if (LN == 3) { FL <- GList; }
  
  # Figure out image dimensions based on first image
  writeLines(sprintf('Importing image data:\n%s',FL[1]));
  HDRinfo <- f.read.nifti.header(FL[1]);
  imsize <- HDRinfo$dim[2:5];

  # Figure out how much padding is needed for each dim  
  pdsize <- 256; 
  
  # pdsize <- 2^nextpow2(max(imsize));
  pds <- pdsize - imsize;

  # Calculate number of unique points (aka Nyquist Limit)
  NUP = ceil((pdsize + 1) / 2);
  
  # Sampling Rate = 1 vox : 1 mm^3
  Fs = 1;
  
  #  Frequency domain (axis)
  FD <- (c(0:(NUP-1)) * Fs) / pdsize;

  # Use anatomical image to mask jacobians?
  if (masking == T) {
    ANATfn <- '/yourPath/symTemplate.nii';                               # EDIT HERE
    msk <- f.read.nifti.volume(ANATfn);
    msk <- 0+(msk > 0);
  }

  # Perform statistics and save results to text files
  if (calc_stats == 1) { 
    
    # Calculate for each subject
    for (ss in 1:length(FL)) {  
    
         # Placeholder/output file in TMP directory if it doesn't already exist
       if (masking == T) { dpfn <- sprintf('%s/MSK_FFTsub%04d.txt',tmpdir,ss); }
       if (masking == F) { dpfn <- sprintf('%s/FFTsub%04d.txt',tmpdir,ss); }
       if (file.exists(dpfn)==F) { 
        
          # Create placeholder & announce progress
          write.table(c(0,0,0), file=dpfn, sep ="\t", col.names=FALSE, row.names=FALSE, qmethod="double");
          writeLines(sprintf('%s  Saved placeholder file: %s ...',Sys.time(),dpfn));
          writeLines(sprintf('Processing subject %04d (%04d total)',ss,length(FL)));

          # Import subject data
          img <- f.read.nifti.volume(FL[ss]);
 
          # Mask jacobian input?
          if (masking == T) { img <- img * msk; }

          # Create a table for output
          frq <- FD; cdx <- 0*FD; cdy <- 0*FD; cdz <- 0*FD;
          csdf <- data.frame(frq,cdx,cdy,cdz);

          # For each dimension (e.g., x or 1)
          vknt <- rep(NA,3);
          for (dd in 1:3) {
            od <- setdiff(c(1:3),dd); # other dims (e.g., y+z or 2:3)
            vknt[dd] <- imsize[od[1]] * imsize[od[2]];
            for (aa in 1:imsize[od[1]]) {
              for (bb in 1:imsize[od[2]]) {
                 # Grab vector (x) at position y, z
                 if (dd == 1) { wndx <- ',%d,%d,1'; } 
                 # Grab vector (y) at position x, z
                 if (dd == 2) { wndx <- '%d,,%d,1'; } 
                 # Grab vector (z) at position x, y
                 if (dd == 3) { wndx <- '%d,%d,,1'; }
                 cmmd <- sprintf('vc <- img[%s]',sprintf(wndx,aa,bb));
                 eval(parse(text=cmmd))
                 # Zero-pad the vector for speed, precision
                 vc <- c(vc,rep(0,pds[dd]));
                 # Perform FFT on the unidimensional vector
                 fvc <- fft(vc);
                 # fft output is symmetrical around MAXIMUM (reflects)
                 # Erase upper half / copy-reflection
                 fvc = fvc[1:NUP];
                 # Magnitude of fft of x and scale fft so it is length-independent
                 mx = abs(fvc)/length(vc);
                 mx = 2 * (mx ^ 2); 
                 dmx <- 20 * log10(mx+1); # pow2db(mx+1)
                 # Plot(FD,dmx,type='l');
                 # Save cumulative counts for x, y, z:
                 if (dd == 1) { csdf$cdx <- csdf$cdx+dmx; } 
                 if (dd == 2) { csdf$cdy <- csdf$cdy+dmx; } 
                 if (dd == 3) { csdf$cdz <- csdf$cdz+dmx; }                  
              }
            }
          }
          # Compute average distributions too
          csdf$AVx <- csdf$cdx / vknt[1];
          csdf$AVy <- csdf$cdy / vknt[2];
          csdf$AVz <- csdf$cdz / vknt[3];

          # Plot results
          yL <- c(0,max(max(csdf$cdx),max(csdf$cdy),max(csdf$cdz)));
          plot(FD,csdf$cdx,ylim=yL,type='l',col='blue'); par(new=T);
          plot(FD,csdf$cdy,ylim=yL,type='l',col='red');  par(new=T);
          plot(FD,csdf$cdz,ylim=yL,type='l',col='green');

          # Save results table to textfile
          write.table(csdf, file=dpfn, sep ="\t", col.names=FALSE, row.names=FALSE, qmethod="double");
          writeLines(sprintf('%s  Saved results from subject %04d to file: %s ...',Sys.time(),ss,dpfn));

       } # If (file.exists(dpfn)==F)
    } # calculate for each subject  
  } # Perform statistics and save results to text files

}
writeLines('PROGRAM FINISHED')
