% threeStepWarp.m
% Developed By: Kenneth I. Vaden, Jr. and Mark A. Eckert
% Edited: Mark A. Eckert, 2/25/2019

% SUMMARY: This script uses gentler deformations compared to buildSymmetricalTemplate.m 
% to limit anatomical distortion and obtain normalization parameters with increasingly finger 
% resolution with each successive normalization step. Here the normalization target or template
% is the optimal group template created with buildSymmetricalTemplate.m. This three step 
% normalization procedure is designed to produce 3 different spatial scales of warping parameters. 
% These different spatial scales can be valideated with the accompanying spatialResolution_FFT.R script.  

% Please note that we have indicated below where you should provide your 
% project-specific information with % EDIT HERE.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% PARAMETERS
maxstep = 3; % Number of normalization steps

% subject list
sublist = '/yourPath/IDs.txt';                                 % EDIT HERE. subject IDs in nii filename
Isbj = textread(sublist,'%s','delimiter','\n');
Inum = length(Isbj);
numfiles = length(Isbj);

% Input images
Idir = '/yourPath/images';                                     % EDIT HERE. input dir   
Iprf = ''; %                                                   % EDIT HERE. add prefix before subject ID, if necessary (e.g., 'r' for rigidly aligned images) 
Iext = ''; %                                                   % EDIT HERE. extension after subject ID 
Iname = fullfile(Idir,sprintf('%s%%s%s',Iprf,Iext));           % input name (with Subject ID %%s space)
Inamenii = fullfile(Idir,sprintf('%s%%s%s',Iprf,Iext));        % input name (with Subject ID %%s space)
Odir = '/yourPath/images';                                     % EDIT HERE. output dir for warp files at each step

% Check that the images exist
if (Inum==0); error('subject list is empty or does not exist!'); end;
Fnum = 0; for (snum = Isbj'); if(exist(sprintf(Inamenii,snum{1}))); Fnum = Fnum + 1; end; end;
if (Fnum ~= Inum); error(sprintf('number of images (%d) does not match list (%d)!',Fnum,Inum)); else fprintf('Found all %d images\n',Fnum); end;

% Target filename
tdir = '/yourPath/'                                            % EDIT HERE.
tgt = fullfile(tdir,'symAverage_Final_Step11.nii.gz');

% Check that the target file exists
if(~exist(tgt,'file')); error(sprintf('file not found: %s',tgt)); end;

% Create output directories (if not found)
for cstep = 1:maxstep;
    Cdir = fullfile(Odir,sprintf('Step%02d_Xforms',cstep));
    if (~exist(Cdir,'dir')); cmmd = sprintf('mkdir %s',Cdir); disp(cmmd); unix(cmmd); end;
end

% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% FOR EACH STEP
for cstep = 1:maxstep; % current step
        
    % Use ANTS to estimate warps for next image to be processed
    for (snum = Isbj')
        
        % subject filename - root for generated filenames
        sNAM = sprintf('%s%s',snum{1}, Iext);

        % affine/placeholder name - advance past completed steps if
        % system crashes, or new sessions are opened, etc.
        Cdir = fullfile(Odir,sprintf('Step%02d_Xforms',cstep));
        Oname = fullfile(Cdir,sprintf('Step%02d_%sAffine.txt',cstep,sNAM));
        
        % If the file does not already exist, then this sessions gets to work on it
        if (~exist(Oname,'file'));
            
            % Write a textfile-placeholder
            dlmwrite(Oname,0,'delimiter','\t');
            fprintf('\n\n********************************************************\n');
            fprintf('  >> Process %s - placeholder created: %s\n',snum{1},Oname);
            fprintf('********************************************************\n\n');
            
            % ANTS: Calculate warps from alligned individual -> template
            % (2010/11/01) increasing the normalization strength.
            if (cstep==1);
                Bdir = Odir; inpf = fullfile(Bdir, sprintf('%s.nii.gz',sNAM));
                MM = sprintf('-m PR[%s,%s,1,4]',tgt,inpf);
                II='-i 50x1x1'; RR='-r Gauss[3,3]'; TT='-t SyN[1.0]';
            elseif (cstep==2)
                Bdir = fullfile(Odir,sprintf('Step%02d_Xforms',cstep-1));
                inpf = fullfile(Bdir,sprintf('Step%02d_%s.nii.gz',cstep-1,sNAM));
                MM = sprintf('-m PR[%s,%s,1,2]',tgt,inpf);
                II='-i 1x50x1'; RR='-r Gauss[2,2]'; TT='-t SyN[0.75]';
            elseif (cstep==3)
                Bdir = fullfile(Odir,sprintf('Step%02d_Xforms',cstep-1));
                inpf = fullfile(Bdir,sprintf('Step%02d_%s.nii.gz',cstep-1,sNAM));
                MM = sprintf('-m PR[%s,%s,1,2]',tgt,inpf);
                II='-i 1x1x40'; RR='-r Gauss[1,1]'; TT='-t SyN[0.50]';
            end
            OO = sprintf('-o %s',fullfile(Cdir,sprintf('Step%02d_%s.nii.gz',cstep,sNAM)));
            %GG = '--geodesic 1';
            clear cmmd; cmmd = sprintf('time ANTS 3 %s %s %s %s %s %s --go-faster true ',MM,OO,II,RR,TT);
            fprintf('%s\n',cmmd); unix(cmmd); disp(' ');
            
            % Combine steps: list affine and diffeomorphic transforms
            XFORMS = '';
            for cs = (1:cstep);
                CSdir = fullfile(Odir,sprintf('Step%02d_Xforms',cs));
                AF = fullfile(CSdir,sprintf('Step%02d_%sAffine.txt',cs,sNAM));
                DF = fullfile(CSdir,sprintf('Step%02d_%sWarp.nii.gz',cs,sNAM));
                XFORMS=sprintf('%s %s %s',DF,AF,XFORMS);
            end
            
            % Input and warp output file names
            inpf = fullfile(Odir,sprintf('%s.nii',sNAM));                                    % EDIT HERE.
            OUTN = fullfile(Cdir,sprintf('Step%02d_%s.nii.gz',cstep,sNAM));
           
            % 3D implement warps -> create individual normalized anatomy
            clear cmmd; cmmd = sprintf('WarpImageMultiTransform 3 %s %s -R %s %s',inpf,OUTN,tgt,XFORMS);
            fprintf('%s\n',cmmd); unix(cmmd); disp(' ');
            
            % Announce the conclusion of each subject & step
            fprintf('\n\n************************************************************************');
            fprintf('\n STEP %d FINISHED (SUBJECT %s) \n',cstep,snum{1});
            fprintf('************************************************************************\n\n');
            
        else
            fprintf('  >> Skip subject %s - file exists: %s\n',snum{1},Oname)
        end
    end
    
    % Announce the conclusion of each step across subjects
    fprintf('\n\n************************************************************************');
    fprintf('\n STEP %d FINISHED \n',cstep);
    fprintf('************************************************************************\n\n');
    
end

% % % % % % % % % % % % % % % %
% Create final template based on multi-step algorithm
Fnum = 0;  % check: do final warped images exist?
for (snum = Isbj'); 
    Cdir = fullfile(Odir,sprintf('Step%02d_Xforms',maxstep));
    OUTN = fullfile(Cdir,sprintf('Step%02d_%s.nii.gz',maxstep,sNAM));
    if(exist(sprintf(Iname,snum{1}))); Fnum = Fnum + 1; end;
end
if (Fnum < Inum); fprintf('fewer final images (%d) than contained in subject list (%d).\n another session will create template.\n\n ',Fnum,Inum); end;
if (Fnum == Inum); 
    fprintf('final images (%d) = subject list length (%d).\n Current session will create template.\n\n ',Fnum,Inum); 
    OUTN = fullfile(Cdir,sprintf('Step%02d_%s*.nii.gz',maxstep,Iprf));
    OUTT = fullfile(Odir,'avgTemplate.nii');
    cmmd = sprintf('AverageImages 3 %s 0 %s',OUTT,OUTN);
    fprintf('%s\n',cmmd); unix(cmmd); disp(' ');
end;
fprintf('\n********************\nProgram completed.\n********************\n');
